@@ -4,7 +4,9 @@ import asyncio
 import logging
 import openai
 import os
+import re
 import sys
+import traceback
 
 from collections import deque
 from rich.progress import Progress
@@ -50,22 +52,31 @@ async def get_comments(programming_language: str, func_name: str, check: bool, t
         if check:
             content = f'Report ALL comments in the given {programming_language}code that are inconsistent with what the code actually does. Put that report inline as a new comment to the code with an explanation prefaced by `# INCONSISTENT COMMENT`. ONLY RETURN THE UPDATED FUNCTION AND COMMENTS: {the_code}'
         else:
-            content = f'Add comments to the following {programming_language}code. Add both low-level and high-level explanatory comments as per-line comments starting with #, PEP 257 docstrings, and PEP 484 style type annotations. Make sure to add comments before all loops, branches, and complicated lines of code. Infer what each function does, using the names, comments, and computations as hints. If there are existing comments or types, augment them rather than replacing them. If existing comments are inconsistent with the code, correct them. Every function argument and return value should be typed. {translate_text}ONLY RETURN THE UPDATED FUNCTION: {the_code}'
+            content = f'Add comments to all code. Add both low-level and high-level explanatory comments as per-line comments starting with #, PEP 257 docstrings, and PEP 484 style type annotations. Make sure to add comments before all loops, branches, and complicated lines of code. Infer what each function does, using the names, comments, and computations as hints. If there are existing comments or types, augment them rather than replacing them. DO NOT DELETE EXISTING COMMENTS. If existing comments are inconsistent with the code, correct them. Every function argument and return value should be typed. {translate_text}ONLY RETURN THE UPDATED FUNCTION. The code:\n{the_code}'
     elif programming_language == "C" or programming_language == "C++":
         content = f"Add comments to the following {programming_language}code. Add both low-level and high-level explanatory comments as per-line comments. Use Google's comment style. Make sure to add comments before all loops, branches, and complicated lines of code. Infer what each function does, using the names, comments, and computations as hints. If there are existing comments, augment them rather than replacing them. If existing comments are inconsistent with the code, correct them. Use swear words judiciously. {translate_text}ONLY RETURN THE UPDATED FUNCTION: {the_code}"
     else:
         content = f'Add comments to the following {programming_language}code. Add both low-level and high-level explanatory comments as per-line comments. Make sure to add comments before all loops, branches, and complicated lines of code. Infer what each function does, using the names, comments, and computations as hints. If there are existing comments, augment them rather than replacing them. If existing comments are inconsistent with the code, correct them. {translate_text}ONLY RETURN THE UPDATED FUNCTION: {the_code}'
+        logging.info(content)
+        
     try:
         max_trials = 3
         for trial in range(max_trials):
-            completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload={'model': 'gpt-3.5-turbo', 'messages': [{'role': 'system', 'content': 'You are a {programming_language}programming assistant who ONLY responds with blocks of commented and typed code. You never respond with text. Just code, starting with ``` and ending with ```.', 'role': 'user', 'content': content}]})
-            code_block = extract_code_block(completion.json())
+            completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload = { "model": 'gpt-4', "messages" : [{'role': 'system', 'content': 'You are an expert {programming_language}programming assistant who ONLY responds with blocks of commented and typed code. You never respond with text. Just code, starting with ``` and ending with ```.', 'role': 'user', 'content': content}] })
+            # completion = openai.ChatCompletion.create(request_timeout=30, model = 'gpt-4', messages = [{'role': 'system', 'content': 'You are an expert {programming_language}programming assistant who ONLY responds with code.', 'role': 'user', 'content': content}])
+
+            code_block = completion.json()['choices'][0]['message']['content']
+            # code_block = completion['choices'][0]['message']['content']
+            
+            logging.info(code_block)
+            
             if check:
                 if "INCONSISTENT" in code_block:
                     print(f"inconsistency found:\n{code_block}")
                 break
                 
             logging.info(f'PROCESSING {code_block}')
+           
             if validated(the_code, code_block):
                 logging.info(f'Validated code block:\n-----\n{code_block}\n-----')
                 global successful_comments
@@ -81,8 +92,9 @@ async def get_comments(programming_language: str, func_name: str, check: bool, t
                 stripped_code_block = strip_imports.strip_imports(ast.parse(stripped_code_block))
                 if stripped_the_code == stripped_code_block:
                     logging.info(f"COMMENTS EQUAL\n=====\n{stripped_the_code}\n=====\n{stripped_code_block}")
-                    pass
+                    break
                 else:
+                    continue
                     logging.info(f"COMMENTS NOT EQUAL\n=====\n{stripped_the_code}\n=====\n{stripped_code_block}")
                     # Otherwise, just splice in the types and the
                     # docstring from the generated function into the
@@ -108,6 +120,9 @@ async def get_comments(programming_language: str, func_name: str, check: bool, t
         import sys
         sys.exit(1)
     except Exception as e:
+        print(f"Commentator exception: {e}")
+        print(f"Please post as an issue to https://github.com/plasma-umass/commentator")
+        traceback.print_exc()
         return ''
     progress.update(pbar, advance=1)
     return code_block
@@ -286,54 +301,38 @@ def now_has_types(code1, code2):
         remove_annotations(node)
     return ast.unparse(tree1) != ast.unparse(tree2)
 
-class ExtractFunction(ast.NodeVisitor):
-    def __init__(self, name):
-        self.names = []
-        self.current_class = deque()
-        self.name = name
-        self.the_function = None
+class FunctionExtractor(ast.NodeVisitor):
+    def __init__(self, target: str):
+        self.target = target.split('.')
+        self.current = []
 
-    def visit_ClassDef(self, node):
-        self.current_class.appendleft(node.name)
+    def visit_FunctionDef(self, node: ast.FunctionDef):
+        self.current.append(node.name)
+        if self.current == self.target:
+            self.result = node
+            return
         self.generic_visit(node)
-        self.current_class.popleft()
-        
-    def visit_FunctionDef(self, node):
-        if not self.the_function:
-            self.process_function(node)
-    def visit_AsyncFunctionDef(self, node):
-        if not self.the_function:
-            self.process_function(node)
-    def process_function(self, node):
-        if len(self.current_class) > 0:
-            name = self.current_class[0] + '.' + node.name
-        else:
-            name = node.name
-        if self.name == name:
-            self.the_function = node
-        
+        self.current.pop()
+
+    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
+        self.current.append(node.name)
+        if self.current == self.target:
+            self.result = node
+            return
+        self.generic_visit(node)
+        self.current.pop()
+
+    def visit_ClassDef(self, node: ast.ClassDef):
+        self.current.append(node.name)
+        self.generic_visit(node)
+        self.current.pop()
+
 def extract_function_ast(program_str: str, function_name: str) -> Union[ast.FunctionDef, ast.AsyncFunctionDef]:
-    """
-    Extract the abstract syntax tree (AST) for a function with a given name from a given program string.
+    module = ast.parse(program_str)
+    function_extractor = FunctionExtractor(function_name)
+    function_extractor.visit(module)
+    return function_extractor.result
 
-    Args:
-        program_str (str): A string representing the program code.
-        function_name (str): A string representing the name of the function to extract the AST for.
-
-    Returns:
-        ast.FunctionDef: The AST node representing the function definition.
-
-    Raises:
-        ValueError: If no function with the given name is found in the AST.
-    """
-    program_ast = ast.parse(program_str)
-    xf = ExtractFunction(function_name)
-    xf.visit(program_ast)
-    function_node = xf.the_function
-    # function_node = next((n for n in program_ast.body if isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef)) and n.name == function_name), None)
-    if function_node is None:
-        raise ValueError(f"No function named '{function_name}' was found")
-    return function_node
 
 def extract_function_source(program_str, function_name):
     return ast.unparse(extract_function_ast(program_str, function_name))
@@ -343,26 +342,59 @@ class EnumerateFunctions(ast.NodeVisitor):
     def __init__(self):
         self.names = []
         self.current_class = deque()
+        self.current_function = deque()
 
     def visit_ClassDef(self, node):
-        self.current_class.appendleft(node.name)
+        self.current_class.append(node.name)
         self.generic_visit(node)
-        self.current_class.popleft()
+        self.current_class.pop()
         
     def visit_FunctionDef(self, node):
+        self.current_function.append(node.name)
         self.process_function(node)
+        self.current_function.pop()
+        
     def visit_AsyncFunctionDef(self, node):
+        self.current_function.append(node.name)
         self.process_function(node)
+        self.current_function.pop()
         
     def process_function(self, node):
         if len(self.current_class) > 0:
-            name = self.current_class[0] + '.' + node.name
+            name = '.'.join(self.current_class) # + '.' + node.name
+        elif len(self.current_function) > 1:
+            name = '.'.join(self.current_function) # + '.' + node.name
         else:
             name = node.name
         self.names.append(name)
         self.generic_visit(node)
         
     
+import ast
+from typing import List
+
+class FunctionEnumerator(ast.NodeVisitor):
+    def __init__(self):
+        self.function_names = []
+        self.namespace = []
+
+    def visit_FunctionDef(self, node: ast.FunctionDef):
+        self.namespace.append(node.name)
+        self.function_names.append(".".join(self.namespace))
+        self.generic_visit(node)
+        self.namespace.pop()
+
+    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
+        self.namespace.append(node.name)
+        self.function_names.append(".".join(self.namespace))
+        self.generic_visit(node)
+        self.namespace.pop()
+
+    def visit_ClassDef(self, node: ast.ClassDef):
+        self.namespace.append(node.name)
+        self.generic_visit(node)
+        self.namespace.pop()
+
 def enumerate_functions(program_str: str) -> List[str]:
     """
     Returns a list of names of functions and async functions defined in a given Python program string.
@@ -374,39 +406,52 @@ def enumerate_functions(program_str: str) -> List[str]:
         A list of names of functions and async functions defined in the program.
     """
     try:
-        program_ast = ast.parse(program_str)
+        module = ast.parse(program_str)
+        function_enumerator = FunctionEnumerator()
+        function_enumerator.visit(module)
+        return function_enumerator.function_names
     except SyntaxError:
-        # Failed to parse.
+        # Failed parse
         return []
-    ef = EnumerateFunctions()
-    ef.visit(program_ast)
-    return ef.names
-#    
-#    names = [n.name for n in program_ast.body if isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef))]
-#    return names
+
+import ast
+from typing import List, Union
+
+class FunctionReplacer(ast.NodeTransformer):
+    def __init__(self, target: str, new_node: ast.AST):
+        self.target = target.split('.')
+        self.current = []
+        self.new_node = new_node
+
+    def visit_FunctionDef(self, node: ast.FunctionDef):
+        self.current.append(node.name)
+        if self.current == self.target:
+            return self.new_node
+        result = self.generic_visit(node)
+        self.current.pop()
+        return result
+
+    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
+        self.current.append(node.name)
+        if self.current == self.target:
+            return self.new_node
+        result = self.generic_visit(node)
+        self.current.pop()
+        return result
+
+    def visit_ClassDef(self, node: ast.ClassDef):
+        self.current.append(node.name)
+        result = self.generic_visit(node)
+        self.current.pop()
+        return result
 
 def replace_function(program_str: str, function_name: str, new_function_str: str) -> str:
-    """Replace a function within a Python program with a new function.
-
-    Args:
-        program_str: A string representing a Python program.
-        function_name: The name of the function to be replaced.
-        new_function_str: A string representing the new function to replace the old one.
-
-    Returns:
-        A string representing the modified Python program.
-    """
-    program_ast = ast.parse(program_str)
-    xf = ExtractFunction(function_name)
-    xf.visit(program_ast)
-    function_node = xf.the_function
-    if function_node is None:
-        raise ValueError(f"No function named '{function_name}' was found")
-    new_function_ast = extract_function_ast(new_function_str, function_name)
-    function_node.body = new_function_ast.body
-    function_node.returns = new_function_ast.returns
-    update_args(function_node, new_function_ast)
-    return ast.unparse(program_ast)
+    module = ast.parse(program_str)
+    new_function_node = ast.parse(new_function_str).body[0]
+    function_replacer = FunctionReplacer(function_name, new_function_node)
+    new_module = function_replacer.visit(module)
+    return ast.unparse(new_module)
+    
 
 def extract_names(ast_node: ast.AST) -> Set[str]:
     """
@@ -460,16 +505,23 @@ def find_code_start(code: str) -> int:
     i = 0
     while i < len(lines) and lines[i].strip() == '':
         i += 1
-    first_line = lines[i].strip()
+    start_line = i
+    while i < len(lines) and not lines[i].strip().startswith('```'):
+        i += 1
+    if i >= len(lines):
+        i = start_line
+    first_line = lines[i]
+    offset = 3
     if first_line == '```':
-        return 3
-    if first_line.startswith('```'):
-        word = first_line[3:].strip()
-        if len(word) > 0 and ' ' not in word:
-            return len(word) + 3
+        return offset
+    matched = re.match(r'^```(?P<language>[a-z]*)', first_line)
+    if matched:
+        offset += len(matched.group('language')) + 1
+        word = first_line[offset:].strip()
+        if len(word) >= 0 and ' ' not in word:
+            return len(word) + offset
     return -1
-    '\n    Returns -1 if code block is not found.\n    '
-test = '\n```python\ndef abs(n):\n    # Check if integer is negative\n    if n < 0:\n        # Return the opposite sign of n (i.e., multiply n by -1)\n        return -n\n    else:\n        # Return n (which is already a positive integer or zero)\n        return n\n```\n'
+
 
 def extract_code_block(completion: dict) -> str:
     """
